/** @file
	Parser: @b string class.

	Copyright (c) 2001-2012 Art. Lebedev Studio (http://www.artlebedev.com)
	Author: Alexandr Petrosian <paf@design.ru> (http://paf.design.ru)
*/

#include "pa_vvoid.h"
#include "pa_vfile.h"

volatile const char * IDENT_PA_VVOID_C="$Id: pa_vvoid.C,v 1.6 2012-05-07 20:05:10 moko Exp $" IDENT_PA_VVOID_H;

#ifdef STRICT_VARS
	bool VVoid::strict_vars = false;
#endif
