/* pa_config_fixed.h. */
/* Configure file for autoconf-disabled platforms  */

#define HAVE_CTYPE_H 1
#define HAVE_ERRNO_H 1
#define HAVE_MEMCPY 1
#define HAVE_MEMMOVE 1
#define HAVE_MEMORY_H 1
#define HAVE_STDIO_H 1
#define HAVE_STDLIB_H 1
#define HAVE_STRCHR 1
#define HAVE_STRCMP 1
#define HAVE_STRINGS_H 1
#define HAVE_STRING_H 1
#define HAVE_STRRCHR 1
#define HAVE_SYS_STAT_H 1
#define HAVE_SYS_TYPES_H 1

#define error_t int
#define LT_OBJDIR ".libs/"
#define LT_MODULE_EXT ".dll"
